# -*- coding: utf-8 -*-
from ..errors import *
from flask import Response, redirect
import re

__all__ = (
    "Authenticator",
    "NotAuthenticated"
)

# IMPORTANT: This is provisional code. Might be changed or removed.
#

class NotAuthenticated(Exception):
    pass


class Authenticator(object):
    def authenticate(self, request):
        raise NotImplementedError

    def info_dict(self, request):
        return { 'username' : self.authenticate(request) }

    def logout(self, request, identity):
        return "logged out"


class AbstractBasicAuthenticator(Authenticator):
    def __init__(self, realm=None):
        self.realm = realm or "Default"
        self.pattern = re.compile(r"^(http(?:s?)://)([^/]+.*)$", re.IGNORECASE)

    def logout(self, request, identity):
        headers = {"WWW-Authenticate": 'Basic realm="%s"' % self.realm}
        url_root = request.args.get('url', request.url_root)
        m = self.pattern.search(url_root)
        if m:
            url_root = m.group(1) + "__logout__@" + m.group(2)
            return redirect(url_root, code=302)
        else:
            return Response("logged out", status=401, headers=headers)

class AdminAdminAuthenticator(AbstractBasicAuthenticator):
    """Simple HTTP Basic authenticator for testing purposes. User name and
    password have to be the same. User name is passed as the authenticated
    identity."""
    def __init__(self, realm=None, **options):
        super(AdminAdminAuthenticator, self).__init__(realm=realm)

    def authenticate(self, request):
        auth = request.authorization
        if auth and auth.username == auth.password:
            return auth.username
        else:
            raise NotAuthenticated

        raise NotAuthenticated


class PassParameterAuthenticator(Authenticator):
    """Permissive authenticator that passes an URL parameter (default
    ``api_key``) as idenity."""
    def __init__(self, parameter=None, **options):
        super(PassParameterAuthenticator, self).__init__(**options)
        self.parameter_name = parameter or "api_key"

    def authenticate(self, request):
        return request.args.get(self.parameter_name)


class HTTPBasicProxyAuthenticator(AbstractBasicAuthenticator):
    def __init__(self, realm=None, **options):
        super(HTTPBasicProxyAuthenticator, self).__init__(realm=realm)
        self.realm = realm or "Default"
        self.pattern = re.compile(r"^(http(?:s?)://)([^/]+.*)$", re.IGNORECASE)

    def authenticate(self, request):
        """Permissive authenticator using HTTP Basic authentication that
        assumes the server to be behind a proxy, and that the proxy authenticated the user. 
        Does not check for a password, just passes the `username` as identity"""
        auth = request.authorization

        if auth:
            return auth.username

        raise NotAuthenticated(realm=self.realm)




# -*- encoding: utf-8 -*-

from __future__ import absolute_import

import os.path
from collections import defaultdict
from .cells import Cell, cut_from_string, cut_from_dict, PointCut
from .model import string_to_dimension_level
from .errors import UserError, ConfigurationError, NoSuchDimensionError
from .common import read_json_file, sorted_dependencies
from . import compat

__all__ = (
    "Authorizer",
    "SimpleAuthorizer",
    "AuthorizationError",
    "NotAuthorized",
    "right_from_dict"
)

ALL_CUBES_WILDCARD = '*'

class AuthorizationError(UserError):
    """Raised when there is any authorization-related error. Use
    more specific `NotAuthorized` when access right is denied."""
    pass

class NotAuthorized(AuthorizationError):
    """Raised when user is not authorized for the request."""
    # Note: This is not called NotAuthorizedError as it is not in fact an
    # error, it is just type of signal.

class Authorizer(object):
    def authorize(self, token, cubes):
        """Returns list of authorized cubes from `cubes`. If none of the cubes
        are authorized an empty list is returned.

        Default implementation returs the same `cubes` list as provided.
        """
        return cubes

    def restricted_cell(self, token, cube, cell=None):
        """Restricts the `cell` for `cube` according to authorization by
        `token`. If no cell is provided or the cell is empty then returns
        the restriction cell. If there is no restriction, returns the original
        `cell` if provided or `None`.
        """
        return cell

    def hierarchy_limits(self, token, cube):
        """Returns a list of tuples: (`dimension`, `hierarchy`, `level`)."""
        # TODO: provisional feature, might change
        return []


class NoopAuthorizer(Authorizer):
    def __init__(self):
        super(NoopAuthorizer, self).__init__()


class _SimpleAccessRight(object):
    def __init__(self, roles, allowed_cubes, denied_cubes, cell_restrictions,
                 hierarchy_limits):
        self.roles = set(roles) if roles else set([])
        self.cell_restrictions = cell_restrictions or {}

        self.hierarchy_limits = defaultdict(list)

        if hierarchy_limits:
            for cube, limits in hierarchy_limits.items():
                for limit in limits:
                    if isinstance(limit, compat.string_type):
                        limit = string_to_dimension_level(limit)
                    self.hierarchy_limits[cube].append(limit)

        self.hierarchy_limits = dict(self.hierarchy_limits)

        self.allowed_cubes = set(allowed_cubes) if allowed_cubes else set([])
        self.denied_cubes = set(denied_cubes) if denied_cubes else set([])
        self._get_patterns()

    def _get_patterns(self):
        self.allowed_cube_suffix = []
        self.allowed_cube_prefix = []
        self.denied_cube_suffix = []
        self.denied_cube_prefix = []

        for cube in self.allowed_cubes:
            if cube.startswith("*"):
                self.allowed_cube_suffix.append(cube[1:])
            if cube.endswith("*"):
                self.allowed_cube_prefix.append(cube[:-1])

        for cube in self.denied_cubes:
            if cube.startswith("*"):
                self.denied_cube_suffix.append(cube[1:])
            if cube.endswith("*"):
                self.denied_cube_prefix.append(cube[:-1])

    def merge(self, other):
        """Merge `right` with the receiver:

        * `allowed_cubes` are merged (union)
        * `denied_cubes` are merged (union)
        * `cell_restrictions` from `other` with same cube replace restrictions
          from the receiver"""

        self.roles |= other.roles
        self.allowed_cubes |= other.allowed_cubes
        self.denied_cubes |= other.denied_cubes

        for cube, restrictions in other.cell_restrictions.items():
            if not cube in self.cell_restrictions:
                self.cell_restrictions[cube] = restrictions
            else:
                self.cell_restrictions[cube] += restrictions

        for cube, limits  in other.hierarchy_limits.items():
            if not cube in self.hierarchy_limits:
                self.hierarchy_limits[cube] = limits
            else:
                self.hierarchy_limits[cube] += limits

        self._get_patterns()

    def is_allowed(self, name, allow_after_denied=True):

        allow = False
        if self.allowed_cubes:
            if (name in self.allowed_cubes) or \
                        (ALL_CUBES_WILDCARD in self.allowed_cubes):
                allow = True

            if not allow and self.allowed_cube_prefix:
                allow = any(name.startswith(p) for p in self.allowed_cube_prefix)
            if not allow and self.allowed_cube_suffix:
                allow = any(name.endswith(p) for p in self.allowed_cube_suffix)

        deny = False
        if self.denied_cubes:
            if (name in self.denied_cubes) or \
                        (ALL_CUBES_WILDCARD in self.denied_cubes):
                deny = True

            if not deny and self.denied_cube_prefix:
                deny = any(name.startswith(p) for p in self.denied_cube_prefix)
            if not deny and self.denied_cube_suffix:
                deny = any(name.endswith(p) for p in self.denied_cube_suffix)

        """
        Four cases:
            - allow match, no deny match
              * allow_deny: allowed
              * deny_allow: allowed
            - no allow match, deny match
              * allow_deny: denied
              * deny_allow: denied
            - no match in either
              * allow_deny: denied
              * deny_allow: allowed
            - match in both
              * allow_deny: denied
              * deny_allow: allowed
        """

        # deny_allow
        if allow_after_denied:
            return allow or not deny
        # allow_deny
        else:
            return allow and not deny

    def to_dict(self):
        as_dict = {
            "roles": list(self.roles),
            "allowed_cubes": list(self.allowed_cubes),
            "denied_cubes": list(self.denied_cubes),
            "cell_restrictions": self.cell_restrictions,
            "hierarchy_limits": self.hierarchy_limits
        }

        return as_dict


def right_from_dict(info):
    return _SimpleAccessRight(
        roles=info.get('roles'),
        allowed_cubes=info.get('allowed_cubes'),
        denied_cubes=info.get('denied_cubes'),
        cell_restrictions=info.get('cell_restrictions'),
        hierarchy_limits=info.get('hierarchy_limits')
    )

class SimpleAuthorizer(Authorizer):
    __options__ = [
        {
            "name": "rights_file",
            "description": "JSON file with access rights",
            "type": "string"
        },
        {
            "name": "roles_file",
            "description": "JSON file with access right roles",
            "type": "string"
        },
        {
            "name": "order",
            "description": "Order of allow/deny",
            "type": "string",
            "values": ["allow_deny", "deny_allow"]
        },
        {
            "name": "guest",
            "description": "Name of the 'guest' role",
            "type": "string",
        },

    ]

    def __init__(self, rights_file=None, roles_file=None, roles=None,
                 rights=None, identity_dimension=None, order=None,
                 guest=None, **options):
        """Creates a simple JSON-file based authorizer. Reads data from
        `rights_file` and `roles_file` and merge them with `roles` and
        `rights` dictionaries respectively."""

        super(SimpleAuthorizer, self).__init__()

        roles = roles or {}
        rights = rights or {}

        if roles_file and not os.path.isabs(roles_file):
            roles_file = os.path.join(options["cubes_root"], roles_file)

        if rights_file and not os.path.isabs(rights_file):
            rights_file = os.path.join(options["cubes_root"], rights_file)

        if roles_file:
            content = read_json_file(roles_file, "access roles")
            roles.update(content)

        if rights_file:
            content = read_json_file(rights_file, "access rights")
            rights.update(content)

        self.roles = {}
        self.rights = {}
        self.guest = guest or None

        order = order or "deny_allow"

        if order == "allow_deny":
            self.allow_after_denied = False
        elif order == "deny_allow":
            self.allow_after_denied = True
        else:
            raise ConfigurationError("Unknown allow/deny order: %s" % order)

        # Process the roles
        for key, info in roles.items():
            role = right_from_dict(info)
            self.roles[key] = role

        deps = dict((name, role.roles) for name, role in self.roles.items())
        order = sorted_dependencies(deps)

        for name in order:
            role = self.roles[name]
            for parent_name in role.roles:
                parent = self.roles[parent_name]
                role.merge(parent)

        # Process rights
        for key, info in rights.items():
            right = right_from_dict(info)
            self.rights[key] = right

            for role_name in list(right.roles):
                role = self.roles[role_name]
                right.merge(role)

        if identity_dimension:
            if isinstance(identity_dimension, compat.string_type):
                (dim, hier, _) = string_to_dimension_level(identity_dimension)
            else:
                (dim, hier) = identity_dimension[:2]
            self.identity_dimension = dim
            self.identity_hierarchy = hier
        else:
            self.identity_dimension = None
            self.identity_hierarchy = None

    def expand_roles(self, info):
        """Merge `right` with its roles. `right` has to be a dictionary.
        """
        right = right_from_dict(info)
        for role_name in list(right.roles):
            role = self.roles[role_name]
            right.merge(role)

        return right

    def right(self, token):
        try:
            right = self.rights[token]
        except KeyError:
            # if guest role exists, use it
            if self.guest and self.guest in self.roles:
                return self.roles[self.guest]
            else:
                raise NotAuthorized("Unknown access right '%s'" % token)

        return right

    def authorize(self, token, cubes):
        try:
            right = self.right(token)
        except NotAuthorized:
            return []

        authorized = []

        for cube in cubes:
            cube_name = str(cube)

            if right.is_allowed(cube_name, self.allow_after_denied):
                authorized.append(cube)

        return authorized

    def restricted_cell(self, identity, cube, cell):
        right = self.right(identity)

        cuts = right.cell_restrictions.get(cube.name)

        # Append cuts for "any cube"
        any_cuts = right.cell_restrictions.get(ALL_CUBES_WILDCARD, [])
        if any_cuts:
            cuts += any_cuts

        if cuts:
            restriction_cuts = []
            for cut in cuts:
                if isinstance(cut, compat.string_type):
                    cut = cut_from_string(cut, cube)
                else:
                    cut = cut_from_dict(cut)
                cut.hidden = True
                restriction_cuts.append(cut)

            restriction = Cell(cube, restriction_cuts)
        else:
            restriction = Cell(cube)

        ident_dim = None
        if self.identity_dimension:
            try:
                ident_dim = cube.dimension(self.identity_dimension)
            except NoSuchDimensionError:
                # If cube has the dimension, then use it, otherwise just
                # ignore it
                pass

        if ident_dim:
            hier = ident_dim.hierarchy(self.identity_hierarchy)

            if len(hier) != 1:
                raise ConfigurationError("Identity hierarchy has to be flat "
                                         "(%s in dimension %s is not)"
                                         % (str(hier), str(ident_dim)))

            # TODO: set as hidden
            cut = PointCut(ident_dim, [identity], hierarchy=hier, hidden=True)
            restriction = restriction & Cell(cube, [cut])

        if cell:
            return cell & restriction
        else:
            return restriction

    def hierarchy_limits(self, token, cube):
        right = self.right(token)

        return right.hierarchy_limits.get(str(cube), [])





# -*- coding=utf -*-
"""SQL Expression compiler"""

# The compiler is meant to be maintained in a similar way as the star schema
# generator is – is to remain as much Cubes-independent as possible, just be a
# low level module somewhere between SQLAlchemy and Cubes.

from __future__ import absolute_import

import sqlalchemy.sql as sql

from expressions import Compiler
from .functions import get_aggregate_function

from ..errors import ExpressionError


__all__ = [
    "SQLExpressionContext",
    "compile_attributes",
    "SQLExpressionCompiler",
]


SQL_FUNCTIONS = [
    # String
    "lower", "upper", "left", "right", "substr",
    "lpad", "rpad", "replace",
    "concat", "repeat", "position",

    # Math
    "round", "trunc", "floor", "ceil",
    "mod", "remainder",
    "sign",

    "min", "max",

    "pow", "exp", "log", "log10",
    "sqrt",
    "cos", "sin", "tan",

    # Date/time
    "extract",

    # Conditionals
    "coalesce", "nullif", "case",

]

# TODO: Add: lstrip, rstrip, strip -> trim
# TODO: Add: like

SQL_AGGREGATE_FUNCTIONS = [
    "sum", "min", "max", "avg", "stddev", "variance", "count"
]

SQL_ALL_FUNCTIONS = SQL_FUNCTIONS + SQL_AGGREGATE_FUNCTIONS;

SQL_VARIABLES = [
    "current_date", "current_time", "local_date", "local_time"
]


class SQLExpressionContext(object):
    """Context used for building a list of all columns to be used within a
    single SQL query."""

    def __init__(self, columns=None, parameters=None, label=None):
        """Creates a SQL expression compiler context.

        * `bases` is a dictionary of base columns or column expressions
        * `for_aggregate` is a flag where `True` means that the expression is
          expected to be an aggregate expression
        * `label` is just informative context label to be used for debugging
          purposes or in an exception. Can be a cube name or a dimension
          name.
        """

        if columns:
            self._columns = dict(columns)
        else:
            self._columns = {}
        self.parameters = parameters or {}
        self.label = label

    @property
    def columns(self):
        return self._columns

    def resolve(self, variable):
        """Resolve `variable` – return either a column, variable from a
        dictionary or a SQL constant (in that order)."""

        if variable in self._columns:
            return self._columns[variable]

        elif variable in self.parameters:
            result = self.parameters[variable]

        elif variable in SQL_VARIABLES:
            result = getattr(sql.func, variable)()

        else:
            label = " in {}".format(self.label) if self.label else ""
            raise ExpressionError("Unknown attribute, variable or parameter "
                                  "'{}'{}" .format(variable, label))

        return result

    def __getitem__(self, item):
        return self.resolve(item)

    def function(self, name):
        """Return a SQL function"""
        if name not in SQL_ALL_FUNCTIONS:
            raise ExpressionError("Unknown function '{}'"
                                  .format(name))
        return getattr(sql.func, name)

    def add_column(self, name, column):
        self._columns[name] = column


def compile_attributes(bases, dependants, parameters, coalesce=None,
                       label=None):
    """Compile dependant attributes in `dependants`. `bases` is a dictionary
    of base attributes and their column expressions."""

    context = SQLExpressionContext(bases, parameters, label=label)
    compiler = SQLExpressionCompiler()

    for attr in dependants:
        # TODO: remove this hasattr with something nicer
        if hasattr(attr, "function") and attr.function:
            # Assumption: only aggregates have function, no measures or other
            # attributes (important!)
            #
            # Aggregation function names are case in-sensitive.
            #
            # If `coalesce_measure` is `True` then selected measure column is
            # wrapped in ``COALESCE(column, 0)``.

            function_name = attr.function.lower()
            function = get_aggregate_function(function_name)
            column = function(attr, context, coalesce)
        else:
            column = compiler.compile(attr.expression, context)

        context.add_column(attr.ref, column)

    return context.columns


class SQLExpressionCompiler(Compiler):
    def __init__(self, context=None):
        super(SQLExpressionCompiler, self).__init__(context)

    def compile_literal(self, context, literal):
        return sql.expression.bindparam("literal",
                                        literal,
                                        unique=True)

    def compile_binary(self, context, operator, op1, op2):
        if operator == "*":
            result = op1 * op2
        elif operator == "/":
            result = op1 / op2
        elif operator == "%":
            result = op1 % op2
        elif operator == "+":
            result = op1 + op2
        elif operator == "-":
            result = op1 - op2
        elif operator == "&":
            result = op1 & op2
        elif operator == "|":
            result = op1 | op2
        elif operator == "<":
            result = op1 < op2
        elif operator == "<=":
            result = op1 <= op2
        elif operator == ">":
            result = op1 > op2
        elif operator == ">=":
            result = op1 >= op2
        elif operator == "=":
            result = op1 == op2
        elif operator == "!=":
            result = op1 != op2
        elif operator == "and":
            result = sql.expression.and_(op1, op2)
        elif operator == "or":
            result = sql.expression.or_(op1, op2)
        else:
            raise SyntaxError("Unknown operator '%s'" % operator)

        return result

    def compile_variable(self, context, variable):
        name = variable.name
        result = context.resolve(name)
        return result

    def compile_unary(self, context, operator, operand):
        if operator == "-":
            result = (- operand)
        elif operator == "+":
            result = (+ operand)
        elif operator == "~":
            result = (~ operand)
        elif operator == "not":
            result = sql.expression.not_(operand)
        else:
            raise SyntaxError("Unknown unary operator '%s'" % operator)

        return result

    def compile_function(self, context, func, args):
        func = context.function(func.name)
        return func(*args)

